// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: hpcxx_rtsImpl.C,v 1.5 1999/07/22 20:56:57 sdiwan Exp $
#define HPCXX_MAX_CHILDREN 100
// Standard includes
#include <unistd.h>
#include <iostream.h>
#include <libgen.h>
#include <pthread.h>
#include <globus_nexus.h>
#include <string.h>

// HPC++ includes
#include <HPCxx_Profile.h>
#include <hpcxx_rts.h>
#include <hpcxx_rtsImpl.h>

#ifdef TULIP_FORTRAN
#include <HPCxx_Prototype.h>
#endif

// Globus includes
#include <globus_toolkit.h>
#include <globus_nexus.h>
#include <globus_duroc_runtime.h>
#include <globus_duroc_bootstrap.h>
#include <globus_common.h>

// create a global map for mapping function prototypes
//HPCxx_IdProtoMap hpcxx_prototypeMap;


int user_arg_ = 0;
HPCxx_Context *hpcxx_currentContext;
HPCxx_ContextID *hpcxx_currentContextID = NULL;
static int hpcxx_currentPosition = -1;
static globus_nexus_startpoint_t hpcxx_startpoint_master; 
static globus_nexus_endpoint_t hpcxx_endpoint_master;
static int hpcxx_globus_nexus_numNodes;
static bool hpcxx_globus_nexus_is_master_context = false;
static bool hpcxx_is_agency = false;
static HPCxx_ContextID *hpcxx_contextIDs;
static bool gs_already_init=false;
static bool hpcxx_context_connected=false;
static HPCxx_Mutex *GFHandlerMutex;  //SMD


globus_nexus_endpoint_t *hpcxx_get_ep(){ return &hpcxx_endpoint_master; }
globus_nexus_startpoint_t *hpcxx_get_sp(){ return &hpcxx_startpoint_master;}

static const char* hpcxx_basename(const char* path)
{
  if (path == NULL) return NULL;
  else if ((strrchr(path, '/')) == NULL) return path;
  else return (strrchr(path, '/') + 1);
}

//
// This table *must* be created in the same order as the consts
// declared in the header file!!!
//
globus_nexus_handler_t hpcxx_globus_nexusHandlerTable[HPCXX_NUM_HANDLERS] = {
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_invokeGFHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_invokeMFHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_readHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_writeHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_ackHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_returnHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_barrierHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_barrierOutHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_lc_bufferMoveHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_rc_bufferMoveHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)hpcxx_p_bufferMoveHandler},
  {GLOBUS_NEXUS_HANDLER_TYPE_THREADED,
   (globus_nexus_handler_func_t)receive_handler},
};

int globus_nexus_master_context()
{
  return(hpcxx_globus_nexus_is_master_context);
}
// special globus_nexus barrier code.

void hpxx_send_unblock(globus_nexus_startpoint_t *n, int i)
{
  TAU_PROFILE("hpcxx_send_unblock()", "void (globus_nexus_startpoint_t *, int)", 
	      TAU_HPCXX);
  globus_nexus_buffer_t buffer;
  globus_nexus_startpoint_t sp;
  globus_nexus_startpoint_copy(&sp, n);
  globus_nexus_buffer_init(&buffer, globus_nexus_sizeof_int(1)+ 
			   globus_nexus_sizeof_startpoint(&sp,1),
			   0);
  globus_nexus_put_int(&buffer, &i, 1);
  globus_nexus_put_startpoint_transfer(&buffer, &sp, 1);
  globus_nexus_send_rsr(&buffer, &hpcxx_contextIDs[i] , HPCXX_BARRIER_IN_HANDLER, 
			GLOBUS_FALSE, GLOBUS_FALSE);
  globus_nexus_buffer_destroy(&buffer);
}


static HPCxx_SyncQ<int> from_parent;
static HPCxx_SyncQ<int> from_children[HPCXX_MAX_CHILDREN];
static globus_nexus_startpoint_t parent;
static int child_id;

static void notify_parent()
{
  TAU_PROFILE("notify_parent()", "void ()", TAU_HPCXX);
  globus_nexus_buffer_t buffer2;
  globus_nexus_buffer_init(&buffer2, globus_nexus_sizeof_int(1), 0);
  globus_nexus_put_int(&buffer2, &child_id, 1);
  //globus_nexus_startpoint_t s;
  //globus_nexus_startpoint_copy(&s, &parent);
  globus_nexus_send_rsr(&buffer2, &parent, HPCXX_BARRIER_OUT_HANDLER,
			GLOBUS_FALSE, GLOBUS_FALSE);
  globus_nexus_buffer_destroy(&buffer2);
}
 
void hpcxx_barrierHandler(globus_nexus_endpoint_t *endpoint,
			  globus_nexus_buffer_t *buffer,
			  globus_bool_t called_from_non_threaded_handler)
{
  TAU_PROFILE("hpcxx_barrierHandler()", 
	      "void (globus_nexus_endpoint_t *, globus_nexus_buffer_t *, globus_bool_t )", TAU_HPCXX);
  globus_nexus_get_int(buffer, &child_id,1);
  globus_nexus_get_startpoint(buffer, &parent, 1);
  from_parent.write(child_id);
  globus_nexus_buffer_destroy(buffer);  // SMD
}

void hpcxx_barrierOutHandler(globus_nexus_endpoint_t *endpoint,
			     globus_nexus_buffer_t *buffer,
			     globus_bool_t called_from_non_threaded_handler)
{
  TAU_PROFILE("hpcxx_barrierOutHandler()", 
	      "void (globus_nexus_endpoint_t *, globus_nexus_buffer_t *, globus_bool_t )", TAU_HPCXX);
  int x;
  globus_nexus_get_int(buffer, &x,1);
  from_children[x].write(x);
  globus_nexus_buffer_destroy(buffer);  // SMD
}

void hpcxx_barrier(){
  TAU_PROFILE("hpcxx_barrier()", "void ()", TAU_HPCXX);
  int i, j;
  if(globus_nexus_master_context()){
    for(i = 1; i < hpcxx_globus_nexus_numNodes; i++) 
      hpxx_send_unblock( hpcxx_currentContextID, i);
    if(hpcxx_globus_nexus_numNodes >=  HPCXX_MAX_CHILDREN)
      printf("too many nodes: HPCXX_MAX_CHILDREN=%d\n", HPCXX_MAX_CHILDREN);
    
    for(i = 1; i < hpcxx_globus_nexus_numNodes; i++) from_children[i].read(j);
    
    for(i = 1; i < hpcxx_globus_nexus_numNodes; i++) 
      hpxx_send_unblock( hpcxx_currentContextID, i);
  }
  else{
    from_parent.read(j);
    notify_parent();
    from_parent.read(j);
  }
}
// end of barrier code.

void hpcxx_typeInit();

void init_thread_key();

/* BEGIN NEW STUFF for FAULT TOLERANCE */

int globus_nexus_fault_callback(void *user_arg, int fault_code) {
  cerr << "globus_nexus got a fatal error, errno = " << fault_code << endl;
  //  cerr << globus_nexus_fault_strings[fault_code] << endl;
  return 0;  // means cleanup and continue
}

/* END NEW STUFF for FAULT TOLERANCE */

HPCxx_ContextID *hpcxx_getMyContextID(void)
{
	return hpcxx_currentContextID;
}      

int hpcxx_initAgent(int &argc, char **&argv)
{
  return hpcxx_initAgent(argc, argv, NULL, NULL);
}

int hpcxx_initAgent(int &argc, char **&argv, int (*fault_callback)(void *user_arg, int code), void *user_arg)
{
  int rc;

  TAU_PROFILE("hpcxx_initAgent()", "int (int, char **)", TAU_HPCXX);
  hpcxx_typeInit();
  init_thread_key();

  if ((rc = gc_init(argv)) != GLOBUS_SUCCESS){
    globus_libc_printf("gc_init failed, rc=%d.\n", rc);
    fflush(NULL);
    return rc;
  }

  // NEW STUFF for FAULT TOLERANCE
  if(fault_callback==NULL) {
  	globus_nexus_enable_fault_tolerance(globus_nexus_fault_callback,
				                        (void *)&user_arg_);
  }
  else {
  	globus_nexus_enable_fault_tolerance(fault_callback, user_arg);
  }
  
  globus_endpoint_setup(&hpcxx_endpoint_master,
			hpcxx_globus_nexusHandlerTable,
			HPCXX_NUM_HANDLERS);
  globus_nexus_startpoint_bind(&hpcxx_startpoint_master,
			       &hpcxx_endpoint_master);

  if((rc=gc_connect(&hpcxx_endpoint_master))!=GLOBUS_SUCCESS) {
    globus_libc_printf("gc_connect failed, rc=%d.\n", rc);
    fflush(NULL);
    return rc;
  }
  
  hpcxx_currentContextID = &hpcxx_startpoint_master;
  
  hpcxx_currentContext =  new HPCxx_Context(hpcxx_currentContextID,
					    true,
					    "child_context");
  hpcxx_is_agency=true;
  hpcxx_currentPosition=0;
  GFHandlerMutex = new HPCxx_Mutex; // SMD
  return GLOBUS_SUCCESS;
}


int hpcxx_init(int &argc, char **&argv, HPCxx_Group *&group)
{
	return hpcxx_init(argc, argv, group, NULL, NULL);
}

int hpcxx_init(int &argc, char **&argv, HPCxx_Group *&group, int (*fault_callback)(void *user_arg, int code), void *user_arg)
{
  int rc;
  
  TAU_PROFILE("hpcxx_init()", "int (int, char **, HPCxx_Group)", TAU_HPCXX);
  TAU_PROFILE_INIT(argc, argv);

  hpcxx_typeInit();
  init_thread_key();

  // This func initializes all of the DUROC constructs we'll need
  // To get started
  if((rc=duroc_init(argc, argv))!=GLOBUS_SUCCESS) {
    globus_libc_printf("Unable to initialize DUROC : %d\n", rc);
    duroc_abort();
  }
  
  // NEW STUFF for FAULT TOLERANCE
  if(fault_callback==NULL) {
  	globus_nexus_enable_fault_tolerance(globus_nexus_fault_callback,
				                        (void *)&user_arg_);
  }
  else {
  	globus_nexus_enable_fault_tolerance(fault_callback, user_arg);
  }
  
  // Here we initialize our endpoint/startpoint pair - note that this
  // used to be done in nexus_boot but that is now gone - we hafta
  // do it ourselves, which is fine 'cause it's easier this way anyway
  globus_endpoint_setup(&hpcxx_endpoint_master,
			hpcxx_globus_nexusHandlerTable,
			HPCXX_NUM_HANDLERS);
  globus_nexus_startpoint_bind(&hpcxx_startpoint_master,
			       &hpcxx_endpoint_master);
  
  // RENDEZVOUS - this takes care of everyone getting their startpoints
  // distributed properly - no need to do it explicitly now!
  if((rc=duroc_rendezvous(hpcxx_startpoint_master,
				(globus_nexus_startpoint_t **)&hpcxx_contextIDs)) != GLOBUS_SUCCESS) {
    globus_libc_printf("Unable to rendezvous : %d\n", rc);
    duroc_abort();
  }

  // If we get here OK then all contexts have their startpoint arrays
  // and global variables set up properly
  hpcxx_globus_nexus_numNodes=getNumProcs();
  hpcxx_currentPosition=getMyID();

  if(hpcxx_currentPosition==0) hpcxx_globus_nexus_is_master_context = true;
  
  hpcxx_currentContext=(HPCxx_Context *) &hpcxx_startpoint_master;
  
  group = new HPCxx_Group(&hpcxx_contextIDs, 
			  hpcxx_globus_nexus_numNodes, 
			  HPCXX_GEN_LOCAL_GROUPID,
			  hpcxx_basename(hpcxx_currentPosition==0 ?
					 argv[0] : "child"));
  group->setLocalIndex(hpcxx_currentPosition);
  hpcxx_currentContext = new HPCxx_Context(hpcxx_currentContextID,
					   hpcxx_currentPosition==0 ?
					   true: false,
					   hpcxx_currentPosition==0 ?
					   "master_context":"child_context");  
  
  GFHandlerMutex = new HPCxx_Mutex;  // SMD

  TAU_PROFILE_SET_NODE(0);
  TAU_PROFILE_SET_CONTEXT(group->getLocalIndex());
  hpcxx_barrier();
  return(HPCXX_SUCCESS);
}


// Thomas Stuckey suggested this ...
int hpcxx_exit()
{
  TAU_PROFILE("hpcxx_exit()", "int ()", TAU_HPCXX);
  if(!hpcxx_is_agency) {
    duroc_cleanup();
  }
  else {
    gc_cleanup();
  }
  globus_nexus_startpoint_destroy(&hpcxx_startpoint_master);
  return(0);
}

int hpcxx_exit(HPCxx_Group*& group)
{
  TAU_PROFILE("hpcxx_exit()", "int (HPCxx_Group *)", TAU_DEFAULT);
  // Only the master context should shutdown globus_nexus
  // the other contexts should simply exit.
  hpcxx_barrier();
  return hpcxx_exit();
}

int hpcxx_connectContext(char *machine)
{
  int rc;

  if(!gs_already_init) {
    if((rc=gs_init())!=GLOBUS_SUCCESS) {
      globus_libc_printf("ERROR: unable to init gs_library : %d\n", rc);
      return NULL;
    }
    else gs_already_init=true;
  }

  if((rc=gs_connect(machine)) != GLOBUS_SUCCESS) {
    globus_libc_printf("ERROR: gs_connect screwed up...dunno how\n");
  }
  hpcxx_context_connected=true;
  return rc;
}

static HPCxx_Mutex createContextMutex;

HPCxx_ContextID *hpcxx_createContext(char *machine, char *exec_path)
{
  globus_nexus_startpoint_t *newSP;
  int rc;
  
  TAU_PROFILE("hpcxx_createContext()", "HPCxx_ContextID * (char *, char *)",
	      TAU_HPCXX);

  createContextMutex.lock();
  
  if(!gs_already_init) {
    if((rc=gs_init())!=GLOBUS_SUCCESS) {
      globus_libc_printf("ERROR: unable to init gs_library : %d\n", rc);
      hpcxx_context_connected=false;

	 createContextMutex.unlock();
	   
      return NULL;
    }
    else gs_already_init=true;
  }
  
  if(!hpcxx_context_connected && (rc=gs_connect(machine)) != GLOBUS_SUCCESS) {
    globus_libc_printf("ERROR: gs_connect screwed up...dunno how\n");
    hpcxx_context_connected=false;

    createContextMutex.unlock();
    
    return NULL;
  }
  
  newSP=gs_create_process(NULL, machine, exec_path);
  
  if(newSP == NULL){
    printf("null node array in createContext\n");
    hpcxx_context_connected=false;

    createContextMutex.unlock();

    return NULL;
  }
  hpcxx_context_connected=false;

  createContextMutex.unlock();

  return newSP;
}

// hpcxx_destroyContext : Destroys a contextID and therefore a startpoint
// BDT 4/23/98 (Happy Anniversary, Wanda!)
// revised 9/23/98 (!) to use new gs_disconnect() func
// re-revised 11/23/98 (!) to not use new gs_disconnect() func
int hpcxx_destroyContextID(HPCxx_ContextID *context)
{
  TAU_PROFILE("hpcxx_destroyContextID()", "int (HPCxx_ContextID *)", TAU_HPCXX);
  int retval;

  if(context==NULL) return 0;
  //retval=gs_disconnect((globus_nexus_startpoint_t *)context);
  retval=globus_nexus_startpoint_destroy((globus_nexus_startpoint_t *)context);
  return retval;
}


void (*userCallback)(void *)=NULL;
void *userCallbackArg=NULL;

// hpcxx_AllowAttach : This function must be called in order to allow
// other HPC++ contexts to attach to a running context which calls it.
// The port argument is used if it contains a non - null value, otherwise
// a port will be chosen from the available ports at random.  You probably
// don't want this, since you need to know which port you are using
// in order to have other contexts attach to it...
int hpcxx_allowAttach(unsigned short attach_port, void (*callback)(void *), void *arg)
{
  TAU_PROFILE("hpcxx_allowAttach()", 
    "int (unsigned short, void (*)(void *), void *)", TAU_HPCXX);
  char *attach_host;
  int retval;
 
  retval = globus_nexus_allow_attach(&attach_port, &attach_host, attach_requested, NULL);
  if(retval) return -1;
  else {
    userCallback=callback;
    userCallbackArg=arg;
    return (int)attach_port;
  }
}

// attach_requested : This func handles the binding of the remote
// startpoint to the local endpoint.  Returns 0 if successful, non-
// zero otherwise, though what these non-zero values mean is beyond
// me at this point, as there's no documentation for this stuff anywhere. 8(
static int attach_requested(void *arg, char *url, globus_nexus_startpoint_t *sp)
{
  TAU_PROFILE("attach_requested()", 
    "int (void *, char *, globus_nexus_startpoint_t *)", TAU_HPCXX);
  int retval;

  retval=globus_nexus_startpoint_bind(sp, &hpcxx_endpoint_master);
  if(userCallback) userCallback(userCallbackArg);
  return retval;
}

// hpcxx_grabContext : makes an attempt to attach to the context given
// by a host name and port number.  This information is concatenated
// together to form a string of the following form:
//
// x-nexus://host:port
//
// Return a new HPCxx_ContextID (which is actually just a globus_nexus_startpoint_t)
// on success, NULL on failure.
int hpcxx_grabContext(HPCxx_ContextID *newID, char *host, int port)
{
  TAU_PROFILE("hpcxx_grabContext()", 
    "HPCxx_ContextID * (HPCxx_ContextID *newID, char *, int)", TAU_HPCXX);
  int retval;
  char xurl[256];
  globus_nexus_startpoint_t *sp=(globus_nexus_startpoint_t *)newID;

  sprintf(xurl, "x-nexus://%s:%d", host, port);
  retval=globus_nexus_attach(xurl, sp);
  if(retval) {
    //globus_nexus_startpoint_destroy(newID);
  }
  else {
	//globus_nexus_startpoint_copy(localID, newID);
	//globus_nexus_startpoint_destroy(newID);
  }
  return retval;
}

//void hpcxx_poll()
//{
//  TAU_PROFILE("hpcxx_poll()", "void ()", TAU_HPCXX);
//  globus_nexus_poll();
//}

void *hpcxx_malloc(size_t size)
{
  TAU_PROFILE("hpcxx_malloc()", "void * (size_t)", TAU_HPCXX);
  return(globus_malloc(size));
}

void hpcxx_free(void *ptr)
{
  TAU_PROFILE("hpcxx_free()", "void (void *)", TAU_HPCXX);
  if (ptr) globus_free(ptr);
}

// Return the number of nodes in this computation
int hpcxx_getNumNodes()
{
  TAU_PROFILE("hpcxx_getNumNodes()", "int ()", TAU_HPCXX);
  return(hpcxx_globus_nexus_numNodes);
}

// Get a list of all nodes
// User must 'delete []' the null terminated node array
HPCxx_GlobalPtr<HPCxx_Node> *hpcxx_getNodes()
{
  // FIXME - not implemented
  return((HPCxx_GlobalPtr<HPCxx_Node> *)0);
}


// Print out an error in the master context only
void hpcxx_msg(const char *msg)
{
  globus_libc_printf((char *)msg);
}

// Abort the hpcxx computation.  hpcxx_abort() will call hpcxx_exit()
void hpcxx_abort(int)
{
  if(hpcxx_is_agency) {
    gc_abort();
  }
  else {
    duroc_abort();
  }
}

bool hpcxx_contextid_equal(HPCxx_ContextID *id1, HPCxx_ContextID *id2)
{
  return((bool)globus_nexus_startpoint_equal(id1, id2));
}


// two prototypes that should be in an include.
int hpcxx_runRF(void * gp, void *p, int funcID, HPCxx_Buffer &b);
int hpcxx_runResult(void * sync, int funcID, HPCxx_Buffer &b);
int hpcxx_runRead(void * sync, void *q, int typeID, int count, int offset);
int hpcxx_runWrite(void * sync, void *q, int typeID, int count, int offset,
		   HPCxx_Buffer &b);

// Globus_Nexus Handler for Remote (global) Function invocation

void hpcxx_invokeGFHandler(globus_nexus_endpoint_t *endpoint,
                           globus_nexus_buffer_t *buffer,
                           globus_bool_t called_from_non_threaded_handler)
{
  TAU_PROFILE("hpcxx_invokeGFHandler()", 
	      "void (globus_nexus_endpoint_t *, globus_nexus_buffer_t *, globus_bool_t)", TAU_HPCXX);
  int funcID, offset;
  globus_nexus_startpoint_t reply_done_sp;
  
  // unpack the global pointer (without calling hpcxx_unpack), probably
  // because the type is not known at this point
  GFHandlerMutex->lock(); // SMD
  globus_nexus_get_startpoint(buffer, &reply_done_sp, 1);
  GFHandlerMutex->unlock(); // SMD

  globus_nexus_get_int(buffer, &offset, 1);
  
  globus_nexus_get_int(buffer, &funcID, 1);
  HPCxx_BufferImpl bi(*buffer);
  HPCxx_Buffer b(bi);
  void *q = (void *) globus_nexus_endpoint_get_user_pointer(endpoint);
  // q = q+offset;  this is wrong.  the offset must be added to typed pointer.
  hpcxx_runRF((void *) &reply_done_sp, q,  funcID, b);
  GFHandlerMutex->lock();  // SMD
  globus_nexus_startpoint_destroy(&reply_done_sp);  // SMD
  GFHandlerMutex->unlock();  // SMD
  globus_nexus_buffer_destroy (buffer); // SMD
}

void hpcxx_returnHandler(globus_nexus_endpoint_t *endpoint,
			 globus_nexus_buffer_t *buffer,
			 globus_bool_t called_from_non_threaded_handler)
{
  TAU_PROFILE("hpcxx_returnHandler()", 
	      "void (globus_nexus_endpoint_t *, globus_nexus_buffer_t *, globus_bool_t )", TAU_HPCXX);
  int funcID;
  void *d = (void *) globus_nexus_endpoint_get_user_pointer(endpoint);
  globus_nexus_get_int(buffer, &funcID, 1);
  HPCxx_BufferImpl bi(*buffer);
  HPCxx_Buffer b(bi);
  hpcxx_runResult(d, funcID, b);  
  globus_nexus_buffer_destroy (buffer); // SMD
}

// Globus_Nexus Handler for Remote Member Function invocation
void hpcxx_invokeMFHandler(globus_nexus_endpoint_t *endpoint,
                           globus_nexus_buffer_t *buffer,
                           globus_bool_t called_from_non_threaded_handler)
{
  // FIXME
}

// Globus_Nexus Handler for GlobalPointer reads
void hpcxx_readHandler(globus_nexus_endpoint_t *endpoint,
		       globus_nexus_buffer_t *buffer,
		       globus_bool_t called_from_non_threaded_handler)
{
  TAU_PROFILE("hpcxx_readHandler()", 
	      "void (globus_nexus_endpoint_t *, globus_nexus_buffer_t *, globus_bool_t )", TAU_HPCXX);
  int count, offset;
  int typeID;
  globus_nexus_startpoint_t reply_done_sp;
  globus_nexus_get_startpoint(buffer, &reply_done_sp, 1);
  globus_nexus_get_int(buffer, &typeID, 1);
  globus_nexus_get_int(buffer, &count, 1);
  globus_nexus_get_int(buffer, &offset, 1);
  void *q = (void *) globus_nexus_endpoint_get_user_pointer(endpoint);
  hpcxx_runRead((void *) &reply_done_sp, q, typeID, count, offset);
  globus_nexus_buffer_destroy (buffer); // SMD
}

// Globus_Nexus Handler for GlobalPointer writes
void hpcxx_writeHandler(globus_nexus_endpoint_t *endpoint,
			globus_nexus_buffer_t *buffer,
			globus_bool_t called_from_non_threaded_handler)
{
  TAU_PROFILE("hpcxx_writeHandler()", 
	      "void (globus_nexus_endpoint_t *, globus_nexus_buffer_t *, globus_bool_t )", TAU_HPCXX);
  int count, offset;
  int typeID;
  globus_nexus_startpoint_t reply_done_sp;
  globus_nexus_get_startpoint(buffer, &reply_done_sp, 1);
  globus_nexus_get_int(buffer, &typeID, 1);
  globus_nexus_get_int(buffer, &count, 1);
  globus_nexus_get_int(buffer, &offset, 1);
  HPCxx_BufferImpl bi(*buffer);
  HPCxx_Buffer hpbuff(bi);
  void *q = (void *) globus_nexus_endpoint_get_user_pointer(endpoint);
  hpcxx_runWrite((void *) &reply_done_sp, q, typeID, count, offset, hpbuff);
  globus_nexus_buffer_destroy (buffer); // SMD
}

// Globus_Nexus Handler to acknowledge a received global communication
void hpcxx_ackHandler(globus_nexus_endpoint_t *endpoint,
                      globus_nexus_buffer_t *buffer,
                      globus_bool_t called_from_non_threaded_handler)
{
  // FIXME
}

// the following is a pack routine for a buffer block object
// it is never to be called because a buffer block object should
// not move.  BUT this is needed by the aix C++ compiler.

void hpcxx_pack(HPCxx_Buffer &, HPCxx_BufferBlock *, int){
}

void hpcxx_lc_bufferMoveHandler(globus_nexus_endpoint_t *endpoint,
				globus_nexus_buffer_t *buffer,
				globus_bool_t called_from_non_threaded_handler){
  TAU_PROFILE("hpcxx_lc_bufferMoveHandler()", 
	      "void (globus_nexus_endpoint_t *, globus_nexus_buffer_t *, globus_bool_t )", TAU_HPCXX);
  HPCxx_BufferImpl bi(*buffer);
  HPCxx_Buffer *b = new HPCxx_Buffer(bi);
  HPCxx_BufferBlock * p = 
    (HPCxx_BufferBlock *) globus_nexus_endpoint_get_user_pointer(endpoint);
  p->lcBuffer = b;
  int i = 1;
  (p->lcLock).write(i);
  i = p->lcLock2;
}

void hpcxx_sendToBufferBlockLC(HPCxx_Buffer &b, HPCxx_ContextID &target){
  TAU_PROFILE("hpcxx_sendToBufferBlockLC()", 
	      "void (HPCxx_Buffer, HPCxx_ContextID )", TAU_HPCXX);
  globus_nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_LCBUFF_MOVE_HANDLER,
			GLOBUS_FALSE, GLOBUS_FALSE);
}

void hpcxx_rc_bufferMoveHandler(globus_nexus_endpoint_t *endpoint,
				globus_nexus_buffer_t *buffer,
				globus_bool_t called_from_non_threaded_handler){
  TAU_PROFILE("hpcxx_rc_bufferMoveHandler()", 
	      "void (globus_nexus_endpoint_t *, globus_nexus_buffer_t *, globus_bool_t )", TAU_HPCXX);
  HPCxx_BufferImpl bi(*buffer);
  HPCxx_Buffer *b = new HPCxx_Buffer(bi);
  HPCxx_BufferBlock * p = 
    (HPCxx_BufferBlock *) globus_nexus_endpoint_get_user_pointer(endpoint);
  p->rcBuffer = b;
  int i = 1;
  (p->rcLock).write(i);
  i = p->rcLock2;
  
}

void hpcxx_sendToBufferBlockRC(HPCxx_Buffer &b, HPCxx_ContextID & target){
  TAU_PROFILE("hpcxx_sendToBufferBlockRC()", 
	      "void (HPCxx_Buffer, HPCxx_ContextID )", TAU_HPCXX);
  globus_nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_RCBUFF_MOVE_HANDLER,
			GLOBUS_FALSE, GLOBUS_FALSE);
}

void hpcxx_p_bufferMoveHandler(globus_nexus_endpoint_t *endpoint,
			       globus_nexus_buffer_t *buffer,
			       globus_bool_t called_from_non_threaded_handler){
  TAU_PROFILE("hpcxx_p_bufferMoveHandler()", 
	      "void (globus_nexus_endpoint_t *, globus_nexus_buffer_t *, globus_bool_t )", TAU_HPCXX);
  HPCxx_BufferImpl bi(*buffer);
  HPCxx_Buffer *b = new HPCxx_Buffer(bi);
  HPCxx_BufferBlock * p = 
    (HPCxx_BufferBlock *) globus_nexus_endpoint_get_user_pointer(endpoint);
  p->pBuffer = b;
  int i = 1;
  (p->pLock).write(i);
  i = p->pLock2;
}

void hpcxx_sendToBufferBlockP(HPCxx_Buffer &b,HPCxx_ContextID & target){
  TAU_PROFILE("hpcxx_sendToBufferBlockP()", 
	      "void (HPCxx_Buffer, HPCxx_ContextID )", TAU_HPCXX);
  globus_nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_PBUFF_MOVE_HANDLER,
			GLOBUS_FALSE, GLOBUS_FALSE);
}

void hpcxx_sendGlobalFunction(HPCxx_Buffer &b, HPCxx_ContextID &target){
  TAU_PROFILE("hpcxx_sendGlobalFunction()", 
	      "void (HPCxx_Buffer, HPCxx_ContextID)", TAU_HPCXX);
  globus_nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_INVOKE_GF_HANDLER, 
			GLOBUS_TRUE, GLOBUS_FALSE);
}
void hpcxx_returnGlobalFunction(HPCxx_Buffer &b, void *target){
  TAU_PROFILE("hpcxx_returnGlobalFunction()", "void (HPCxx_Buffer, void *)", 
	      TAU_HPCXX);
  globus_nexus_send_rsr(b.impl_.buffer_, (HPCxx_ContextID *) target,
			HPCXX_RETURN_HANDLER, GLOBUS_TRUE, GLOBUS_FALSE);
}

void receive_handler(globus_nexus_endpoint_t *endpoint,
                     globus_nexus_buffer_t *buffer,
                     globus_bool_t is_non_threaded_handler)
{
}
