// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_BufferImpl.C,v 1.1 1999/04/17 01:00:21 kate Exp $
#include <stdio.h>
#include <HPCxx_BufferImpl.h>

const int HPCXX_NO_BUFFER_SPACE = -1;

HPCxx_BufferImpl::HPCxx_BufferImpl(int bufSize)
{
 buffer_ = new nexus_buffer_t;
 if( nexus_buffer_init(buffer_, bufSize, 0)){
   exit(0);
   }
}

HPCxx_BufferImpl::HPCxx_BufferImpl(nexus_buffer_t &buffer)
  : buffer_(&buffer)
{
}

HPCxx_BufferImpl::~HPCxx_BufferImpl()
{
//  nexus_buffer_destroy(buffer_);
}

int HPCxx_BufferImpl::pack(const HPCxx_ContextID* data, int count)
{
    int i;
    nexus_startpoint_t *theSPs = new nexus_startpoint_t[count];
    for(i = 0; i < count; i++)
        nexus_startpoint_copy(&(theSPs[i]), &( ((HPCxx_ContextID*) data)[i] ));
    int err = nexus_check_buffer_size(buffer_,
				    nexus_sizeof_startpoint(theSPs,count),
				    nexus_sizeof_startpoint(theSPs,count),
				    0,
				    0);
    if (err < 0) {
        printf("pack error\n");
        return(HPCXX_NO_BUFFER_SPACE);
        }
    nexus_put_startpoint_transfer(buffer_, theSPs, count);

  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const float* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
				    nexus_sizeof_float(count),
				    nexus_sizeof_float(count),
				    0,
				    0);
  if (err < 0) {
      printf("pack error\n");
      return(HPCXX_NO_BUFFER_SPACE);
  }
  nexus_put_float(buffer_, (float*) data, count);

  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const double* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
                                    nexus_sizeof_double(count),
                                    nexus_sizeof_double(count),
                                    0,
                                    0);
  if (err < 0) return(HPCXX_NO_BUFFER_SPACE);

  nexus_put_double(buffer_, (double*) data, count);

  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const short* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
                                    nexus_sizeof_short(count),
                                    nexus_sizeof_short(count),
                                    0,
                                    0);
  if (err < 0) return(HPCXX_NO_BUFFER_SPACE);

  nexus_put_short(buffer_, (short*) data, count);

  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const u_short* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
                                    nexus_sizeof_u_short(count),
                                    nexus_sizeof_u_short(count),
                                    0,
                                    0);
  if (err < 0) return(HPCXX_NO_BUFFER_SPACE);

  nexus_put_u_short(buffer_, (u_short*) data, count);
  
  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const int* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
                                    nexus_sizeof_int(count),
                                    nexus_sizeof_int(count),
                                    0,
                                    0);
  if (err < 0) return(HPCXX_NO_BUFFER_SPACE);

  nexus_put_int(buffer_, (int*) data, count);

  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const u_int* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
                                    nexus_sizeof_u_int(count),
                                    nexus_sizeof_u_int(count),
                                    0,
                                    0);
  if (err < 0) return(HPCXX_NO_BUFFER_SPACE);

  nexus_put_u_int(buffer_, (u_int*) data, count);

  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const long* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
                                    nexus_sizeof_long(count),
                                    nexus_sizeof_long(count),
                                    0,
                                    0);
  if (err < 0) return(HPCXX_NO_BUFFER_SPACE);

  nexus_put_long(buffer_, (long*) data, count);

  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const u_long* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
                                    nexus_sizeof_u_long(count),
                                    nexus_sizeof_u_long(count),
                                    0,
                                    0);
  if (err < 0) return(HPCXX_NO_BUFFER_SPACE);

  nexus_put_u_long(buffer_, (u_long*) data, count);

  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const char* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
                                    nexus_sizeof_char(count),
                                    nexus_sizeof_char(count),
                                    0,
                                    0);
  if (err < 0) return(HPCXX_NO_BUFFER_SPACE);

  nexus_put_char(buffer_, (char*) data, count);

  return(HPCXX_SUCCESS);
}

int HPCxx_BufferImpl::pack(const u_char* data, int count)
{
  int err = nexus_check_buffer_size(buffer_,
                                    nexus_sizeof_u_char(count),
                                    nexus_sizeof_u_char(count),
                                    0,
                                    0);
  if (err < 0) return(HPCXX_NO_BUFFER_SPACE);

  nexus_put_u_char(buffer_, (u_char*) data, count);

  return(HPCXX_SUCCESS);
}
void HPCxx_BufferImpl::unpack(HPCxx_ContextID *data, int count)
{
  nexus_get_startpoint(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(float *data, int count)
{
  nexus_get_float(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(double *data, int count)
{
  nexus_get_double(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(short *data, int count)
{
  nexus_get_short(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(u_short *data, int count)
{
  nexus_get_u_short(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(int *data, int count)
{
  nexus_get_int(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(u_int *data, int count)
{
  nexus_get_u_int(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(long *data, int count)
{
  nexus_get_long(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(u_long *data, int count)
{
  nexus_get_u_long(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(char *data, int count)
{
  nexus_get_char(buffer_, data, count);
}

void HPCxx_BufferImpl::unpack(u_char *data, int count)
{
  nexus_get_u_char(buffer_, data, count);
}


