//-----------------------------------------------------------------------------
// hpcxx_invoke.C -- contains function declarations used in invoking remote
// member functions
//-----------------------------------------------------------------------------

#include <HPCxx_Profile.h>
#include <HPCxx_Group.h>
#include <HPCxx_Buffer.h>
#include <hpcxx_invoke.h>
#include <hpcxx_invokeImpl.cc>
#include <HPCxx_Sync.cc>
#include <HPCxx_SyncQ.cc>
#include <HPCxx_GlobalPtr.cc>
#include <HPCxx_GlobalRefImpl.cc>
#include <hpcxx_register.cc>

hpcxx_id_t barrierSetUp_id = 24001;

HPCxx_SyncQ<int> barrier_set_up_lock[2];
HPCxx_GlobalPtr<HPCxx_BufferBlock> *barrier_ptr;


HPCxx_GlobalPtr<HPCxx_BufferBlock>
barrierSetUp(HPCxx_GlobalPtr<HPCxx_BufferBlock> parent, hpcxx_id_t ParIndx)
{
  TAU_PROFILE("barrierSetUp()", 
  "HPCxx_GlobalPtr<HPCxx_BufferBlock> (HPCxx_GlobalPtr<HPCxx_BufferBlock, int)",
    TAU_HPCXX);

  int i = barrier_set_up_lock[0];
  HPCxx_GlobalPtr<HPCxx_BufferBlock> *z = barrier_ptr;
  barrier_ptr = new HPCxx_GlobalPtr<HPCxx_BufferBlock>(parent);
  barrier_set_up_lock[1] = ParIndx; 
  return *z;
}


void barrierRegister(){
  barrierSetUp_id = hpcxx_register(barrierSetUp, barrierSetUp_id); 
}


void hpcxx_typeInit(void) {
  TAU_PROFILE("hpcxx_typeInit()", "void ()", TAU_HPCXX);
  HPCxx_RegisteredType<int> int_registration;
  HPCxx_RegisteredType<u_int> u_int_registration;
  HPCxx_RegisteredType<float> float_registration;
  HPCxx_RegisteredType<double> double_registration;
  HPCxx_RegisteredType<char> char_registration;
  HPCxx_RegisteredType<u_char> u_char_registration;
  HPCxx_RegisteredType<long> long_registration;
  HPCxx_RegisteredType<u_long> u_long_registration;
  HPCxx_RegisteredType<short> short_registration;
  HPCxx_RegisteredType<u_short> u_short_registration;
  barrierRegister();
 // barrierSetUp_id = hpcxx_register(barrierSetUp, barrierSetUp_id); 
}




// Global pointer type specializations.  These really should go in a
// HPCxx_GlobalPtr.C, which doesn't exist.

HPCxx_GenericPtr::~HPCxx_GenericPtr(void)
{
}

#ifdef AIX
hpcxx_type_t HPCxx_GlobalPtr<int>::type(void) const
{
  return HPCXX_TYPE_INT;
}

hpcxx_type_t HPCxx_GlobalPtr<float>::type(void) const
{
  return HPCXX_TYPE_FLOAT;
}

hpcxx_type_t HPCxx_GlobalPtr<double>::type(void) const {
  return HPCXX_TYPE_DOUBLE;
}
#else
template<> hpcxx_type_t HPCxx_GlobalPtr<int>::type(void) const
{
  return HPCXX_TYPE_INT;
}

template<> hpcxx_type_t HPCxx_GlobalPtr<float>::type(void) const
{
  return HPCXX_TYPE_FLOAT;
}

template<> hpcxx_type_t HPCxx_GlobalPtr<double>::type(void) const {
  return HPCXX_TYPE_DOUBLE;
}
#endif


// Force compiler to instantiate (explicitly) those classes that will be
// needed at link stage.

#ifdef AIX

class HPCxx_Sync<int>;
class HPCxx_SyncQ<int>;
class HPCxx_GlobalPtr<int>;
class HPCxx_GlobalPtr<float>;
class HPCxx_GlobalPtr<double>;
class HPCxx_GlobalRefImpl<int>;
class HPCxx_GlobalRefImpl<float>;
class HPCxx_GlobalRefImpl<double>;
class HPCxx_GlobalPtr<HPCxx_BufferBlock>;
class HPCxx_GlobalRefImpl<HPCxx_BufferBlock>;
class HPCxx_Sync<HPCxx_GlobalPtr<HPCxx_BufferBlock> >;

hpcxx_id_t hpcxx_registerClass(int*);
hpcxx_id_t hpcxx_registerClass(unsigned int*);
hpcxx_id_t hpcxx_registerClass(float*);
hpcxx_id_t hpcxx_registerClass(double*);
hpcxx_id_t hpcxx_registerClass(char*);
hpcxx_id_t hpcxx_registerClass(unsigned char*);
hpcxx_id_t hpcxx_registerClass(long*);
hpcxx_id_t hpcxx_registerClass(unsigned long*);
hpcxx_id_t hpcxx_registerClass(short*);
hpcxx_id_t hpcxx_registerClass(unsigned short*);

size_t hpcxx_sizeof(int&);
size_t hpcxx_sizeof(float&);
size_t hpcxx_sizeof(double&);
size_t hpcxx_sizeof(HPCxx_BufferBlock&);

void
hpcxx_pack(HPCxx_Buffer&, HPCxx_GlobalPtr<HPCxx_BufferBlock>*, int);
void
hpcxx_unpack(HPCxx_Buffer&, HPCxx_GlobalPtr<HPCxx_BufferBlock>*, int);

hpcxx_id_t
hpcxx_register(HPCxx_GlobalPtr<HPCxx_BufferBlock>
	       (*)(HPCxx_GlobalPtr<HPCxx_BufferBlock>,int),int);

// For some reason, using the explicit instantiation syntax on linux
// doesn't work, so the following specialization was used instead.

#else

template class HPCxx_Sync<int>;
template class HPCxx_SyncQ<int>;
template class HPCxx_GlobalPtr<int>;
template class HPCxx_GlobalPtr<float>;
template class HPCxx_GlobalPtr<double>;
template class HPCxx_GlobalRefImpl<int>;
template class HPCxx_GlobalRefImpl<float>;
template class HPCxx_GlobalRefImpl<double>;
template class HPCxx_GlobalPtr<HPCxx_BufferBlock>;
template class HPCxx_GlobalRefImpl<HPCxx_BufferBlock>;
template class HPCxx_Sync<HPCxx_GlobalPtr<HPCxx_BufferBlock> >;

template hpcxx_id_t hpcxx_registerClass(int*);
template hpcxx_id_t hpcxx_registerClass(unsigned int*);
template hpcxx_id_t hpcxx_registerClass(float*);
template hpcxx_id_t hpcxx_registerClass(double*);
template hpcxx_id_t hpcxx_registerClass(char*);
template hpcxx_id_t hpcxx_registerClass(unsigned char*);
template hpcxx_id_t hpcxx_registerClass(long*);
template hpcxx_id_t hpcxx_registerClass(unsigned long*);
template hpcxx_id_t hpcxx_registerClass(short*);
template hpcxx_id_t hpcxx_registerClass(unsigned short*);

size_t hpcxx_sizeof(int&);
size_t hpcxx_sizeof(float&);
size_t hpcxx_sizeof(double&);
size_t hpcxx_sizeof(HPCxx_BufferBlock&);

template void
hpcxx_pack(HPCxx_Buffer&, HPCxx_GlobalPtr<HPCxx_BufferBlock>*, int);

template void
hpcxx_unpack(HPCxx_Buffer&, HPCxx_GlobalPtr<HPCxx_BufferBlock>*, int);

template hpcxx_id_t
hpcxx_register(HPCxx_GlobalPtr<HPCxx_BufferBlock>
	       (*)(HPCxx_GlobalPtr<HPCxx_BufferBlock>,int),int);


#endif

int HPCxx_Type<HPCxx_BufferBlock>::registrationID()
{
  TAU_PROFILE("HPCxx_Type<HPCxx_BufferBlock>::registrationID()", "int ()", 
    TAU_HPCXX);
  return HPCxx_BufferBlock::registrationID();
}
