//-----------------------------------------------------------------------------
// To run the program use the following command line arguments:
//      TestRMFI_MInvoke -nx -n 1               #### for nexus
//      mpirun -np 2 TestRMFI_MInvoke           #### for mpi
// Also, it is possible to print out information by using the 
// -DDEBUG compiler switch.
//-----------------------------------------------------------------------------

#include <hpcxx.h>
#include <HPCxx_Profile.h>
#include <math.h>
#include <assert.h>
#include <iostream.h>


// class Bar
// -----------
// This is a simple class which has methods with between zero and six
// parameters.  The methods are used to test the different versions of the 
// hpcxx_minvoke and hpcxx_aminvoke which can call member functions with up to
// six parameters. 
  
class Bar {
private:
  int anInt;
  int id_;

public:
  Bar(int id = 0, int AnInt = 0) : id_(id), anInt(AnInt) { }

  static int registrationID() { return 645; }

  friend void hpcxx_pack (HPCxx_Buffer &buffer, Bar *data, int size = 1);
  friend void hpcxx_unpack (HPCxx_Buffer &buffer, Bar *data, int size = 1);

  int foo0()
  {
    TAU_PROFILE("foo0()", "int ()", TAU_USER);
    return id_;
  }

  int foo1(int bar)
  {
    TAU_PROFILE("foo1()", "int (int )", TAU_USER);
    if (bar == 13)
      return id_;

    cout << "ERROR in execution of function -> foo1\n";
    return 0;
  }

  int foo2(int bar, char ch)
  {
    TAU_PROFILE("foo2()", "int (int, char )", TAU_USER);
    if ((bar == 13) && (ch == 'a'))
      return id_;

    cout << "ERROR in execution of function -> foo2\n";
    return 0;
  }

  int foo3(int bar, char ch, short sh)
  { 
    TAU_PROFILE("foo3()", "int (int, char, short )", TAU_USER);
    if ((bar == 13) && (ch == 'a') && (sh == 41))
      return id_;

    cout << "ERROR in execution of function -> foo3\n";
    return 0;
  }

  int foo4(int bar, char ch, short sh, float fl)
  {
    TAU_PROFILE("foo4()", "int (int, char, short, float )", TAU_USER);
    if ((bar == 13) && (ch == 'a') && (fabs(fl - 2.7182818) < .0000002) &&
	(sh == 41))
      return id_;

    cout << "ERROR in execution of function -> foo4\n";
    return 0;
  }

  int foo5(int bar, char ch, short sh, float fl, long ln)
  {
    TAU_PROFILE("foo5()", "int (int, char, short, float, long )", TAU_USER);
    if ((bar == 13) && (sh == 41) && (fabs(fl - 2.7182818) < .0000002) &&
	(ch == 'a') && (ln == 235463246))
      return id_;
    
    cout << "ERROR in execution of function -> foo5\n";
    return 0;
  }

  int foo6(int bar, char ch, short sh, float fl, long ln, u_long ul)
  {
    TAU_PROFILE("foo5()", "int (int, char, short, float, long, u_long )", 
      TAU_USER);
    if ((bar == 13) && (sh == 41) && (fabs(fl - 2.7182818) < .0000002) &&
	(ch == 'a') && (ln == 235463246) && (ul == 546367))
      return id_;
    
    cout << "ERROR in execution of function -> foo6\n";
    return 0;
  }

};

// Here are the pack and unpack functions for class Bar.

void hpcxx_pack(HPCxx_Buffer &buffer, Bar *data, int size)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer &, Bar *, int )", TAU_USER);
  for (int i=0; i<size; i++) {
    hpcxx_pack(buffer, &data[i].id_);
    hpcxx_pack(buffer, &data[i].anInt);
  }
}


void hpcxx_unpack (HPCxx_Buffer &buffer, Bar *data, int size)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, Bar *, int )", TAU_USER);
  for (int i=0; i<size; i++) {
    hpcxx_unpack(buffer, &data[i].id_);
    hpcxx_unpack(buffer, &data[i].anInt);
  }
}


// A sync variable has been defined to facilitate passing global pointers to
// other contexts.
typedef HPCxx_GlobalPtr<Bar> gptr_to_Bar;
HPCxx_Sync<gptr_to_Bar> a_truly_global_ptr;

// This function passes a global pointer between contexts.
int put_gp(gptr_to_Bar ptr)
{
  TAU_PROFILE("put_gp()", "int (gptr_to_Bar  )", TAU_USER);
  a_truly_global_ptr = ptr;
  return 0;
}

static int result0;
static int result1;
static int result2;

int
main (int argc, char* argv[])
{
  TAU_PROFILE("main()", "int (int, char * )", TAU_DEFAULT);
  HPCxx_Group* group;
  hpcxx_init(argc, argv, group);
  
  int nc = group->getNumContexts() - 1;
  int context = group->getLocalIndex();

  if (nc != 2) {
    cout << "ERROR, you must run with three contexts, either:" << endl;
    cout << "    TestRMFI_Invoke -nx -n 2        ### for nexus, or" << endl;
    cout << "    mpirun -np 3 TestRMFI_Invoke    ### for mpi" << endl;
    hpcxx_exit(group);
    exit(3);
  }

#ifdef DEBUG
  cout << "nc = " << nc << ", context = " << context << endl;
#endif

  // register the member functions
  static int foo_id0 = hpcxx_register(&Bar::foo0, 30);
  static int foo_id1 = hpcxx_register(&Bar::foo1, 31);
  static int foo_id2 = hpcxx_register(&Bar::foo2, 32);
  static int foo_id3 = hpcxx_register(&Bar::foo3, 33);
  static int foo_id4 = hpcxx_register(&Bar::foo4, 34);
  static int foo_id5 = hpcxx_register(&Bar::foo5, 35);
  static int foo_id6 = hpcxx_register(&Bar::foo6, 36);
  static int put_gpid = hpcxx_register(&put_gp, 37);

  int anint = 13;
  char achar = 'a';
  short ashort = 41;
  float afloat = 2.718282;
  long along = 235463246;
  u_long au_long = 546367;
  
#ifdef DEBUG
  cout << "registered functions" << endl;
#endif
  
  hpcxx_barrier();

  if (context == 0) {
    int res;

    Bar* x = new Bar(5, 37);
    HPCxx_GlobalPtr<Bar> Bar_ptr(x);

    Bar* y = new Bar(7, 37);
    HPCxx_GlobalPtr<Bar> Bar_ptr2(y);
  
    // pass the global pointers to another context
    hpcxx_invoke(*group->getContextID(1), res, put_gpid, Bar_ptr);
    hpcxx_invoke(*group->getContextID(2), res, put_gpid, Bar_ptr2);
      
#ifdef DEBUG
    cout << "Sent global pointers to remote contexts" << endl;
#endif

    // simple local test
    hpcxx_invoke(*group->getContextID(0), res, put_gpid, Bar_ptr);
    gptr_to_Bar Bar_gp = a_truly_global_ptr;
    hpcxx_minvoke(&Bar_gp, result0, foo_id0);
    if (result0 == 5) {
      result0 = 0;
    } else {
      cout << "ERROR, minvoke from context 0 fo foo0()" << endl;
    }
  }

  //
  // Test invoke from context 2
  //

  if (context == 2) {
      HPCxx_Sync<int> result[7];
      gptr_to_Bar Bar_gp = a_truly_global_ptr;
      
      hpcxx_aminvoke(&Bar_gp, result[0], foo_id0);
#ifdef DEBUG
      cout << "\n [2]Finished invoke of foo0" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[1], foo_id1, anint);
#ifdef DEBUG
    cout << " [2]Finished invoke of foo1" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[2], foo_id2, anint, achar);
#ifdef DEBUG
    cout << " [2]Finished invoke of foo2" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[3], foo_id3, anint, achar, ashort);
#ifdef DEBUG
    cout << " [2]Finished invoke of foo3" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[4], foo_id4, anint, achar, ashort, 
		     afloat);
#ifdef DEBUG
    cout << " [2]Finished invoke of foo4" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[5], foo_id5, anint, achar, ashort, 
		     afloat, along);
#ifdef DEBUG
    cout << " [2]Finished invoke of foo5" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[6], foo_id6, anint, achar, ashort, 
		     afloat, along, au_long);
#ifdef DEBUG
    cout << " [2]Finished invoke of foo6" << endl;
#endif //DEBUG

      int values[7];
      
      // wait for the methods to finish executing
      for (int k=0; k<7; k++) result[k].read(values[k]);
      
      // Add up the returned values to determine if each method was executed.
      int sum = 0;
      for (int i=0; i<7; i++) sum += values[i];
      
#ifdef DEBUG
    cout << "\nThe result of context 2 is " << sum << endl;
#endif //DEBUG

      // we want to return 0 (for success)
      if (sum == 49) {
	result2 = 0;
      } else {
	cout << "ERROR, results from context 2 failed = " << sum << endl;
      }
    } 

    if (context == 1) {
      
      int result[7];
      gptr_to_Bar Bar_gp = a_truly_global_ptr;

      hpcxx_minvoke(&Bar_gp, result[0], foo_id0);
#ifdef DEBUG
    cout << "\n [1]Finished invoke of foo0 = " << result[0] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[1], foo_id1, anint);
#ifdef DEBUG
    cout << " [1]Finished invoke of foo1 = " << result[1] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[2], foo_id2, anint, achar);
#ifdef DEBUG
    cout << " [1]Finished invoke of foo2 = " << result[2] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[3], foo_id3, anint, achar, ashort);
#ifdef DEBUG
    cout << " [1]Finished invoke of foo3 = " << result[3] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[4], foo_id4, anint, achar, ashort, 
		    afloat);
#ifdef DEBUG
    cout << " [1]Finished invoke of foo4 = " << result[4] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[5], foo_id5, anint, achar, ashort, 
		    afloat, along);
#ifdef DEBUG
    cout << " [1]Finished invoke of foo5 = " << result[5] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[6], foo_id6, anint, achar, ashort, 
		    afloat, along, au_long);
#ifdef DEBUG
    cout << " [1]Finished invoke of foo6 = " << result[6] << endl;
#endif //DEBUG

      // Add up the returned values to determine if each method was executed.
      int sum = 0;
      for (int i=0; i<7; i++) sum += result[i];

#ifdef DEBUG
    cout << "\nThe result of context 1 is " << sum << endl;
#endif //DEBUG

      if (sum == 35) {
	result1 = 0;
      } else {
	cout << "ERROR, results from context 1 failed = " << sum << endl;
      }
    }

  // synchronize execution of the 3 contexts.
  hpcxx_barrier();
  hpcxx_exit(group);
  
#ifdef DEBUG
  cout << "\nContext " << context << " called hpcxx_exit\n";
  cout << "\nThe final result is " << result0 + result1 + result2 << "\n";
#endif //DEBUG

  if (context == 0) {
    cout << endl;
    if (result0 + result1 + result2 != 0) {
      cout << "ERROR, TestRMFI_MInvoke failed = "
	   << result0 + result1 + result2 << endl;
    } else {
      cout << "Successfully completed TestRMFI_MInvoke" << endl;
    }
  }

  return (result0 + result1 + result2);
}

template void hpcxx_unpack(HPCxx_Buffer &, HPCxx_GlobalPtr<Bar> *, int);

