//-----------------------------------------------------------------------------
// HPCxx_Prototype.h -- declarations for a category of classes to support
// function prototypes and to provide a mapping between HPC++ function ids
// and function prototypes.
//
// This file declares two classes:
//     HPCxx_Prototype
//     HPCxx_IdProtoMap
//-----------------------------------------------------------------------------

#ifndef _HPCXX_PROTOTYPE_H_
#define _HPCXX_PROTOTYPE_H_

/*
 * HPCxx_Prototype.h
 */

#include <map.h>
#include <string.h>
#include <iostream.h>
#include <hpcxx_types.h>


//---- class HPCxx_Prototype --------------------------------------------------
// Provides information regarding a function prototypes.
//-----------------------------------------------------------------------------
class HPCxx_Prototype {
 public:
   HPCxx_Prototype();
   HPCxx_Prototype(const char* prototype);
   HPCxx_Prototype(const HPCxx_Prototype& s);
   ~HPCxx_Prototype();

   HPCxx_Prototype& operator=(const HPCxx_Prototype&);

   bool operator==(const HPCxx_Prototype& proto) const
     { return (strcmp(d_prototype, proto.d_prototype) == 0) ? true : false; }

   friend int operator<(const HPCxx_Prototype& s1, const HPCxx_Prototype& s2)
     { return strcmp(s1.d_prototype, s2.d_prototype); }
   
   int numberOfArgs(void);
   hpcxx_type_t argTypeN(int i);
   hpcxx_type_t returnType(void);
   const char* name(void);
   const char* prototype(void) const { return d_prototype; }

   friend ostream& operator<<(ostream& o, const HPCxx_Prototype& s);

 private:
   hpcxx_type_t hpcxx_returnType(char* prototype);
   hpcxx_type_t getArgType(char *typeString);
   
   char* d_prototype;
   char* d_name;
   int d_numArgs;
};


//---- HPCxx_IdProtoMap -------------------------------------------------------
// This class handles the mapping between an HPC++ id and a Prototype object.
//-----------------------------------------------------------------------------
class HPCxx_IdProtoMap {
 public:
   HPCxx_IdProtoMap();
   
   static void createRegisterCode( const char* prototypeFile,
				   const char* codeFile );

   int addPrototype(const HPCxx_Prototype& s, int id = -1);
   int size() const { return d_idProtoMap.size(); }

   HPCxx_Prototype & operator[](int i) { return(d_idProtoMap[i]); }
   int operator[](HPCxx_Prototype& s)  { return(d_protoIdMap[s]); }

 private:
   map<int, HPCxx_Prototype, less<int> > d_idProtoMap;
   map<HPCxx_Prototype, int, less<HPCxx_Prototype> > d_protoIdMap;
   
   int freeID;
};


#endif // HPCXX_PROTOTYPE_H_
